#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <pwd.h>
#include "configfile.h"
#include "options.h"
#include "hiscore.h"
#include "input.h"
#include "terminput.h"
#include "joystick.h"
#include "draw.h"

static const char input_keynames[12][6] = {
	"left", "right", "up", "down", "a", "b", "start",
	"rot", "sdrop", "rot-a", "rot-b", "hdrop"
};
static const char tcolor_keys[8] = "ijlostz";

void writehiscores(FILE *fp);

void cdhome()
{
	struct passwd *pwd;
	char *s = getenv("HOME");
	if (s)
		chdir(s);
	else if (pwd = getpwuid(getuid()))
		chdir(pwd->pw_dir);
}

static int readname(const char *line, char *dest)
{
	int i, n;
	if (line[0] == '[') {
		line++;
		n = 7;
	} else
		n = 11;
	i = 0;
	while (i < n && (i && line[i] == '-' || islower(line[i]) ||
						isdigit(line[i]))) {
		dest[i] = line[i];
		i++;
	}
	if (!i || line[i-1] == '-')
		return 0;
	dest[i] = '\0';
	if (n == 7) {
		if (line[i] != ']')
			return 0;
	} else {
		while (isspace(line[i]))
			i++;
		if (line[i] != '=')
			return 0;
	}
	return 1;
}

static int readopt(const char *line, char *key, union val *val, int *tp)
{
	char *p;
	if (line[0] == '\n')
		return 0;
	if (readname(line, key)) {
		if (line[0] == '[')
			return 1;
#ifndef JOYSTICK
		if (!strcmp(key, "dev"))
			return 0;
#endif
		if (p = strchr(line, '=')) {
			*tp = strtoval(p+1, val);
			return 1;
		}
	}
	return 0;
}

#ifdef JOYSTICK
static void readjsdevname(FILE *fp, char *line)
{
	char *devname = js_devname[sect_hd.next->name[2]!='0'];
	char *p = strchr(line, '=')+1;
	int n;
	sscanf(p, "%s", devname);
	if (!strchr(p, '\n')) {
		fgets(line, 16, fp);
		n = strlen(devname);
		memcpy(devname+n, line, 15-n);
	}
	if (p = strchr(devname, '\n'))
		*p = '\0';
	if (devname[0] != '/' || !strcmp(devname, "/dev"))
		devname[0] = '\0';
}

#  ifdef TWOPLAYER
static void read_inputdev_player(int pl)
{
	int i = 0;
	if (pl != 1 && pl != 2)
		return;
	if (sect_hd.next->name[0] == 'j')
		i = sect_hd.next->name[2]-'0'+1;
	inputdevs_player[i] = pl;
}
#  endif
#endif

static int nametokeypress_i(int i, int flags)
{
	if (i < 6)
		return MVLEFT+i | flags;
#ifdef JOYSTICK
	if (i == 6)
		return STARTBTN;
#endif
	if (i == 12)
		return 0;
	return (i < 11 ? MVUP+i-7 : HARDDROP) | IN_GAME | flags;
}

static int nametokeypress(char *name)
{
	int i = 0;
	int flags;
#ifdef TWOPLAYER
	if (name[0] == 'p') {
		flags = (name[1]=='1') ? PLAYER_1 : PLAYER_2;
		memmove(name, name+2, strlen(name+2)+1);
	} else
#endif
	flags = 0;
	while (strcmp(name, input_keynames[i]) && i < 12)
		i++;
	return nametokeypress_i(i, flags);
}

static int test_input_chr9(const char *s)
{
	int i = 0;
	while (i < 9) {
		if (!strcmp(s, input_chr9[i]))
			return i+1;
		i++;
	}
	return 0;
}

static keytp *stdin_convertval(union val *val, int tp)
{
	keytp *input = (keytp *) val->str;
	if (!tp) {
		if (val->integ <= 0) {
			*input = 0;
			return input;
		} else if (val->integ < 10)
			*input = val->integ+'0';
#ifndef CURSES
		else if (val->integ > ' ')
			return NULL;
	} else if (val->str[0] == ESC) {
		return input;
#endif
	} else if (val->str[1]) {
		*input = test_input_chr9(val->str);
		if (!*input)
			return NULL;
	}
#ifdef CURSES
	else *input = val->str[0];
#else
	input[1] = '\0';
#endif
	return input;
}

static int readkeymapping(char *key, union val val, int tp)
{
	int keypress = nametokeypress(key);
	keytp *input;
	if (!keypress)
		return 0;
#ifdef JOYSTICK
	if (sect_hd.next->name[0] == 'j') {
		if (!tp)
			val.integ += '0';
		else if (!(val.integ = test_input_chr9(val.str)))
			return 0;
		js_setmapping(sect_hd.next->name[2]=='1', val.integ, keypress);
	} else
#endif
      	if ((char) keypress == STARTBTN)
		;
	else if (input = stdin_convertval(&val, tp)) {
		if (*input)
			term_setmapping(input, keypress);
	} else
		return 0;
	return 1;
}

static void read_tetrom_color(const char *key, int val)
{
	char *p;
	if (strlen(key) > 1 || val < 1 || val > 7)
		return;
	if (p = strchr(tcolor_keys, key[0]))
		tetrom_colors[p-tcolor_keys] = val;
}

static int is_inputconf(const char *key)
{
	return !strcmp(key, "stdin")
#ifdef JOYSTICK
		|| !strcmp(key, "js0") || !strcmp(key, "js1")
#endif
		;
}

void readoptions()
{
	FILE *fp;
	char line[16];
	char key[12];
	union val val;
	int tp;
	char inputconf = 0;
	char tcolors = 0;

	if (sect_hd.opts || sect_hd.next)
		return;
	cdhome();
	fp = fopen(CONFIG_FILENAME, "r");
	if (!fp)
		return;
	while (fgets(line, 16, fp)) {
		if (!readopt(line, key, &val, &tp))
			continue;
		if (line[0] == '[') {
			if (!strcmp(key, "hiscore"))
				break;
			inputconf = is_inputconf(key);
			tcolors = !strcmp(key, "tcolors");
			if (!tcolors)
				addsect(key);
			continue;
		}
		if (inputconf) {
#ifdef JOYSTICK
			if (!strcmp(key, "dev") && sect_hd.next->name[0]=='j')
				readjsdevname(fp, line);
			else
#  ifdef TWOPLAYER
			if (!tp && !strcmp(key, "player"))
				read_inputdev_player(val.integ);
			else
#  endif
#endif
			if (!readkeymapping(key, val, tp))
				goto addopt;
			continue;
		} else if (tcolors) {
			read_tetrom_color(key, val.integ);
			continue;
		}
addopt:		addopt(key, val, tp, sect_hd.next);
	}
	fclose(fp);
}

static void writeopts(FILE *fp, const struct sect *sect)
{
	struct option *o = sect->opts;
	char s[5] = "";
	while (o) {
		fprintf(fp, "%s=", opt_key(o));
		if (opt_isint(o))
			fprintf(fp, "%d", o->val.integ);
		else {
			memcpy(s, o->val.str, 4);
			fprintf(fp, "%s", s);
		}
		putc('\n', fp);
		o = o->next;
	}
}

#ifdef JOYSTICK
static void printbtnmapping(FILE *fp, int n, int keypress, int i)
{
	int b = js_getbtnfor(n, keypress);
	if (!b || b == (char) keypress && (!(keypress & IN_GAME) ||
				b == js_getbtnfor(n, keypress ^ IN_GAME)))
		return;
	fprintf(fp, "%s=", input_keynames[i]);
	if (b < 10)
		fprintf(fp, input_chr9[b-1]);
	else
		fprintf(fp, "%d", b-'0');
	putc('\n', fp);
}
#else
#define printbtnmapping(fp, n, k, i)
#endif

static void printkeymapping(FILE *fp, int keypress, int i)
{
	char s[5];
	int c;
#ifndef CURSES
	if (!term_getkeyfor(keypress, s, 0))
		return;
#else
	c = term_getkeyfor_int(keypress, 0);
	if (!c)
		return;
	if (c >= 0x100)
		sprintf(s, "%d", c);
	else {
		s[0] = c;
		s[1] = '\0';
	}
#endif
#ifdef TWOPLAYER
	if (keypress & (PLAYER_1 | PLAYER_2))
		fprintf(fp, "p%c", '1'+!(keypress & PLAYER_1));
#endif
	fprintf(fp, "%s=", input_keynames[i]);
	if (s[0] == '%')
		putc('%', fp);
	else if (s[1] || s[0] > ' ' || s[0] < 0)
		fprintf(fp, s);
	else if (s[0] < 10)
		fprintf(fp, input_chr9[s[0]-1]);
	else
		fprintf(fp, "%d", s[0]);
	putc('\n', fp);
}

static void writeinputconf(FILE *fp, const char *devkey)
{
	int keypress, flags = 0;
	int i = 0;
#ifdef JOYSTICK
#  define isjoystick (devkey[0]=='j')
	if (isjoystick) {
		i = devkey[2]-'0';
		fprintf(fp, "dev=%s\n", js_devname[i]);
		i++;
	}
#  ifdef TWOPLAYER
	if (inputdevs_player[i])
		fprintf(fp, "player=%d\n", inputdevs_player[i]);
#  endif
#else
#  define isjoystick 0
#endif

mappings:
	i = -1;
	do {
		if (i == 5 && !isjoystick || i == 7)
			i += 2;
		else if (i == 11)
			i = 8;
		else
			i++;
		keypress = nametokeypress_i(i, flags);
		if (isjoystick)
			printbtnmapping(fp, devkey[2]=='1', keypress, i);
		else
			printkeymapping(fp, keypress, i);
	} while (i != 8);

	if (isjoystick)
		return;
#ifdef TWOPLAYER
	if (!flags) {
		flags = PLAYER_1;
		goto mappings;
	}
	if (flags & PLAYER_1) {
		flags = PLAYER_2;
		goto mappings;
	}
#endif
}

static void write_tetrom_colors(FILE *fp)
{
	int i;
	if (!strncmp(tetrom_colors, "", 7))
		return;
	fprintf(fp, "[tcolors]\n");
	for (i = 0; i < 7; i++)
		fprintf(fp, "%c=%d\n", tcolor_keys[i], tetrom_colors[i]);
}

int writeconfig()
{
	FILE *fp;
	struct sect *sect;
	readoptions();
	readhiscores(NULL);
	fp = fopen(CONFIG_FILENAME, "w");
	if (!fp)
		return 0;
	writeopts(fp, &sect_hd);
	addsect("stdin");
#ifdef JOYSTICK
	if (*js_devname[0])
		addsect("js0");
	if (*js_devname[1])
		addsect("js1");
#endif
	sect = sect_hd.next;
	while (sect) {
		fprintf(fp, "[%s]\n", sect->name);
		if (is_inputconf(sect->name))
			writeinputconf(fp, sect->name);
		writeopts(fp, sect);
		sect = sect->next;
	}
	freeoptions("");
	write_tetrom_colors(fp);
	if (hiscores[0].score) {
		fprintf(fp, "[hiscore]\n");
		writehiscores(fp);
	}
	fclose(fp);
	return 1;
}
