#include <stdio.h>
#include <unistd.h>
#ifdef CURSES
#include <curses.h>
#else
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#endif
#include "terminput.h"
#include "input.h"

static struct keys {
	keytp menu_keys[6];
	keytp ingame_keys[5];
#ifndef CURSES
	struct escseq {
		char str[3];
		unsigned char keypress;
		struct escseq *next;
	} escseq;
#endif
} keys_singlepl;

#ifdef TWOPLAYER
static struct keys keys_1p, keys_2p;
#endif

#ifndef CURSES
static char unreadchr;

static int readchr()
{
	char c;
	int ret;
	if (!unreadchr) {
		ret = read(STDIN_FILENO, &unreadchr, 1);
		if (ret <= 0)
			return 0;
	}
	c = unreadchr;
	unreadchr = '\0';
	return c;
}

/* wait 10 microseconds for input */
static int readchr_wait()
{
	if (waitinput(STDIN_FILENO, 10))
		return readchr();
	return 0;
}

static int readescseq(char *s)
{
	/* wait 50 ms before returning ESC */
	int c = waitinput(STDIN_FILENO, 50000) ? readchr() : 0;
	if (c == '[') {
		c = readchr_wait();
		if (!c)
			return -1;
		switch (c) {
		case 'A': return MVUP;
		case 'B': return MVDOWN;
		case 'C': return MVRIGHT;
		case 'D': return MVLEFT;
		case '3':
			if ((c = readchr()) == '~')
				return DEL;
			if (c)
				unreadchr = c;
			c = '3';
		}
		s[0] = '[';
		s[1] = c;
		s[2] = readchr_wait();
		if (isdigit(c) && isdigit(s[2]))
			readchr_wait();
	} else {
		if (!c)
			return ESC;
		if (c == ESC) {
			unreadchr = ESC;
			return ESC;
		}
		s[0] = c;
		if (s[1] = readchr_wait())
			s[2] = readchr_wait();
	}
	s[3] = '\0';
	return 0;
}

int term_readkey(char *dest)
{
	int c = readchr();
	if (!c)
		return 0;
	if (c == ESC)
		switch (c = readescseq(dest+1)) {
		case -1:
			return 0;
		case 0:
			dest[0] = ESC;
			return strlen(dest);
		}
	else if (c == DEL)
		c = BACKSPACE;
	dest[0] = c;
	dest[1] = '\0';
	return 1;
}
#else /* if CURSES */
int term_readkey_int()
{
	const short funckeys[7] = {
		KEY_LEFT, KEY_RIGHT, KEY_UP, KEY_DOWN,
		KEY_ENTER, KEY_BACKSPACE, KEY_DC
	};
	const char keys[7] = {
		MVLEFT, MVRIGHT, MVUP, MVDOWN,
		STARTBTN, BACKSPACE, DEL
	};
	int c = getch();
	int i;
	if (c == ERR)
		return 0;
# ifdef KEY_RESIZE
	if (c == KEY_RESIZE)
		return 0;
# endif
	for (i = 0; i < 7; i++) {
		if (c == funckeys[i])
			return keys[i];
	}
	return c;
}
#endif

#ifndef CURSES
static struct escseq *getescseq(int flags)
{
# ifdef TWOPLAYER
	if (flags & PLAYER_1)
		return &keys_1p.escseq;
	else if (flags & PLAYER_2)
		return &keys_2p.escseq;
	else
# endif
	return &keys_singlepl.escseq;
}

static int transl_escseq(const char *s, int flags)
{
	struct escseq *e;
	int ret;
# ifdef TWOPLAYER
	if (!(flags & SINGLE_PL)) {
		if (ret = transl_escseq(s, flags | PLAYER_1 | SINGLE_PL))
			return ret | PLAYER_1;
		if (ret = transl_escseq(s, flags | PLAYER_2 | SINGLE_PL))
			return ret | PLAYER_2;
		return 0;
	}
# endif
	e = getescseq(flags);
	ret = 0;
	do if ((flags & IN_GAME || !(e->keypress & 0x80)) &&
					!strncmp(s, e->str, 3)) {
		ret = e->keypress;
		if (ret & 0x80)
			ret ^= 0x80;
		else if (flags & IN_GAME)
			continue;
		break;
	} while (e = e->next);
	return ret;
}
#endif /* !CURSES */

static int transl_chr_ingame(int c, keytp *keys)
{
	int i = 0;
	keys += keys_singlepl.ingame_keys - keys_singlepl.menu_keys;
	while (i < 5) {
		if (keys[i] == c)
			return i==4 ? HARDDROP : MVUP+i;
		i++;
	}
	return 0;
}

static int transl_chr(int c, int k)
{
	keytp *keys;
	int i;
#ifdef TWOPLAYER
	if (k & PLAYER_1)
		keys = keys_1p.menu_keys;
	else if (k & PLAYER_2)
		keys = keys_2p.menu_keys;
	else if (!(k & SINGLE_PL)) {
		if (i = transl_chr(c, k | PLAYER_1))
			return i | PLAYER_1;
		if (i = transl_chr(c, k | PLAYER_2))
			return i | PLAYER_2;
		return 0;
	} else
#endif
	keys = keys_singlepl.menu_keys;
	if ((k & IN_GAME) && (k = transl_chr_ingame(c, keys)))
		return k;
	for (i = 0; i < 6; i++)
		if (keys[i] == c)
			return MVLEFT+i;
	return 0;
}

int term_getkeypress(int flags)
{
	int c, ret;
#ifndef CURSES
	char s[5];
	ret = term_readkey(s);
	if (!ret)
		return 0;
	c = (unsigned char) s[0];
	if (c == ESC) {
		if (ret == 1)
			return ESC;
		if (ret = transl_escseq(s+1, flags))
			return ret;
		if (!(flags & IN_GAME))
			return ESC+1;
		goto next;
	}
#else
	c = term_readkey_int();
	if (!c || c == ESC)
		return c;
#endif
	if (ret = transl_chr(c, flags))
		return ret;
	if (c <= MVDOWN || !(flags & IN_GAME)) {
#ifdef CURSES
		if (c >= 0x100)
			return ESC+1;
#endif
		return c;
	}
	switch (c) {
	case BACKSPACE:
		return ESC;
	case HARDDROP:
	case STARTBTN:
	case 'p':
	case 'q':
		return c;
	}
#ifndef CURSES
next:	if (waitinput(STDIN_FILENO, 0))
		return term_getkeypress(flags);
#endif
	return 0;
}

static int transl_input(const keytp *input, int flags)
{
#ifndef CURSES
	if (*input == ESC)
		return transl_escseq(input+1, flags);
#endif
	return transl_chr(*input, flags);
}

#ifndef CURSES
static const char *getescseq_str(int keypr)
{
	struct escseq *e = getescseq(keypr);
	unsigned char k = keypr | (keypr & IN_GAME) >> 1;
	do if (k == e->keypress)
		return e->str;
	while (e = e->next);
	return "";
}
#endif

static keytp *getchrfor(int keypress)
{
	keytp *keys;
	int i;
#ifdef TWOPLAYER
	if (keypress & PLAYER_1)
		keys = keys_1p.menu_keys;
	else if (keypress & PLAYER_2)
		keys = keys_2p.menu_keys;
	else
#endif
	keys = keys_singlepl.menu_keys;
	if (keypress & IN_GAME) {
		keys += keys_singlepl.ingame_keys - keys_singlepl.menu_keys;
		if ((keypress & 63) == HARDDROP)
			i = 4;
		else
			i = (keypress & 7)-MVUP;
	} else
		i = (keypress & 7)-MVLEFT;
	return keys+i;
}

static int key_equals(int keypr, const keytp *input)
{
#ifndef CURSES
	if (*input == ESC)
		return !strncmp(input+1, getescseq_str(keypr), 3);
#endif
	return *input == *getchrfor(keypr);
}

static int transl_strict(const keytp *input, int flags)
{
	int k = transl_input(input, flags);
	if (flags & IN_GAME && !key_equals(k | flags, input))
		k = 0;
	return k;
}

static void rmkey(const keytp *input, int keypr)
{
	int flags = keypr & (IN_GAME | PLAYER_1 | PLAYER_2);
	int old;
	if (!(flags & (PLAYER_1 | PLAYER_2)))
		flags |= SINGLE_PL;
	old = transl_strict(input, flags);
	if (old && old != (keypr & 63))
		term_rmmapping(old | flags);
#ifdef TWOPLAYER
	if (flags & SINGLE_PL)
		return;
	flags ^= PLAYER_1 | PLAYER_2;
	if (old = transl_strict(input, flags))
		term_rmmapping(old | flags);
#endif
}

#ifndef CURSES
static void mapescseq(const char *str, int keypr)
{
	struct escseq *e = getescseq(keypr);
	unsigned char k = keypr | (keypr & IN_GAME) >> 1;
	if (e->keypress) {
		while (k != e->keypress) {
			if (!e->next) {
				e = e->next = malloc(sizeof(struct escseq));
				e->next = NULL;
				break;
			}
			e = e->next;
		}
	}
	memcpy(e->str, str, 3);
	e->keypress = k;
}

static void rmescseq(int keypress)
{
	struct escseq *e, *p;
	unsigned char keypr = keypress | (keypress & IN_GAME) >> 1;
	e = getescseq(keypress);
	if (keypr == e->keypress) {
		e->keypress = 0;
		return;
	}
	while (e->next) {
		if (keypr == e->next->keypress) {
			p = e->next->next;
			free(e->next);
			e->next = p;
			break;
		}
		e = e->next;
	}
}
#endif /* !CURSES */

static int ingame_key_equals(int keypr, const keytp *p)
{
	if (keypr & IN_GAME || (keypr & 7) < MVUP)
		return 0;
#ifndef CURSES
	if (!*p) {
		p = getescseq_str(keypr);
		return *p && !strncmp(p, getescseq_str(keypr | IN_GAME), 3);
	}
#endif
	return *p && *p == *getchrfor(keypr | IN_GAME);
}

void term_setmapping(const keytp *input, int keypr)
{
	keytp *c = getchrfor(keypr);
	rmkey(input, keypr);
	if (ingame_key_equals(keypr, c))
		term_setmapping(input, keypr | IN_GAME);
#ifndef CURSES
	if (*input == ESC) {
		*c = 0;
		mapescseq(input+1, keypr);
		return;
	} else if (*getescseq_str(keypr))
		rmescseq(keypr);
#endif
	*c = *input;
}

static int test_fallback(int keypr, const keytp *input)
{
	int flags = keypr & IN_GAME;
	char k = keypr;
#ifdef TWOPLAYER
	int pl = keypr & (PLAYER_1 | PLAYER_2);
	if (!pl)
		flags |= SINGLE_PL;
#endif
	if ((!(keypr & IN_GAME) || k==HARDDROP) && (k==A_BTN || k==B_BTN
#ifdef TWOPLAYER
			|| pl==PLAYER_2 && inputdevs_player[0] != 2
#endif
				))
		return 0;
	keypr = transl_input(input, flags);
	return !keypr ||
#ifdef TWOPLAYER
		pl == (keypr & (PLAYER_1 | PLAYER_2)) &&
#endif
		k == (char) keypr;
}

#ifndef CURSES
int term_getkeyfor(int keypr, char *dest, int fallback)
{
	char c = *getchrfor(keypr);
	const char *p;
	int ret;
	dest[1] = '\0';
	if (c) {
		dest[0] = c;
		return 1;
	}
	p = getescseq_str(keypr);
	if (*p) {
		dest[0] = ESC;
		strncpy(dest+1, p, 3);
		dest[4] = '\0';
		return strlen(dest);
	}
	if (!fallback)
		return 0;
	dest[0] = c = keypr;
	ret = 1;
	if (keypr & IN_GAME && c != HARDDROP)
		ret = term_getkeyfor(keypr ^ IN_GAME, dest, 1);
	if (ret && !test_fallback(keypr, dest))
		ret = 0;
	return ret;
}
#endif

int term_getkeyfor_int(int keypr, int fallback)
{
	keytp c = *getchrfor(keypr);
	char k = keypr;
	if (c)
		return c;
#ifndef CURSES
	if (*getescseq_str(keypr))
		return 0;
#endif
	if (!fallback)
		return 0;
	c = k;
	if (keypr & IN_GAME && k != HARDDROP)
		c = term_getkeyfor_int(keypr ^ IN_GAME, 1);
	if (c && !test_fallback(keypr, &c))
		c = 0;
	return c;
}

void term_rmmapping(int keypress)
{
	keytp *c = getchrfor(keypress);
	*c = 0;
#ifndef CURSES
	rmescseq(keypress);
#endif
}

#ifndef CURSES
static void freeescseq(struct escseq *hd)
{
	struct escseq *e;
	while (e = hd->next) {
		hd->next = e->next;
		free(e);
	}
}

void free_escape_sequences()
{
	freeescseq(&keys_singlepl.escseq);
# ifdef TWOPLAYER
	freeescseq(&keys_1p.escseq);
	freeescseq(&keys_2p.escseq);
# endif
}
#endif
